package aoc2023

import println
import readInput

/*
--- Day 19: Aplenty ---
The Elves of Gear Island are thankful for your help and send you on your way. They even have a hang glider that someone stole from Desert Island; since you're already going that direction, it would help them a lot if you would use it to get down there and return it to them.
As you reach the bottom of the relentless avalanche of machine parts, you discover that they're already forming a formidable heap. Don't worry, though - a group of Elves is already here organizing the parts, and they have a system.
To start, each part is rated in each of four categories:

x: Extremely cool looking
m: Musical (it makes a noise when you hit it)
a: Aerodynamic
s: Shiny

Then, each part is sent through a series of workflows that will ultimately accept or reject the part. Each workflow has a name and contains a list of rules; each rule specifies a condition and where to send the part if the condition is true. The first rule that matches the part being considered is applied immediately, and the part moves on to the destination described by the rule. (The last rule in each workflow has no condition and always applies if reached.)
Consider the workflow ex{x>10:one,m<20:two,a>30:R,A}. This workflow is named ex and contains four rules. If workflow ex were considering a specific part, it would perform the following steps in order:

Rule "x>10:one": If the part's x is more than 10, send the part to the workflow named one.
Rule "m<20:two": Otherwise, if the part's m is less than 20, send the part to the workflow named two.
Rule "a>30:R": Otherwise, if the part's a is more than 30, the part is immediately rejected (R).
Rule "A": Otherwise, because no other rules matched the part, the part is immediately accepted (A).

If a part is sent to another workflow, it immediately switches to the start of that workflow instead and never returns. If a part is accepted (sent to A) or rejected (sent to R), the part immediately stops any further processing.
The system works, but it's not keeping up with the torrent of weird metal shapes. The Elves ask if you can help sort a few parts and give you the list of workflows and some part ratings (your puzzle input). For example:
px{a<2006:qkq,m>2090:A,rfg}
pv{a>1716:R,A}
lnx{m>1548:A,A}
rfg{s<537:gd,x>2440:R,A}
qs{s>3448:A,lnx}
qkq{x<1416:A,crn}
crn{x>2662:A,R}
in{s<1351:px,qqz}
qqz{s>2770:qs,m<1801:hdj,R}
gd{a>3333:R,R}
hdj{m>838:A,pv}

{x=787,m=2655,a=1222,s=2876}
{x=1679,m=44,a=2067,s=496}
{x=2036,m=264,a=79,s=2244}
{x=2461,m=1339,a=466,s=291}
{x=2127,m=1623,a=2188,s=1013}

The workflows are listed first, followed by a blank line, then the ratings of the parts the Elves would like you to sort. All parts begin in the workflow named in. In this example, the five listed parts go through the following workflows:

{x=787,m=2655,a=1222,s=2876}: in -> qqz -> qs -> lnx -> A
{x=1679,m=44,a=2067,s=496}: in -> px -> rfg -> gd -> R
{x=2036,m=264,a=79,s=2244}: in -> qqz -> hdj -> pv -> A
{x=2461,m=1339,a=466,s=291}: in -> px -> qkq -> crn -> R
{x=2127,m=1623,a=2188,s=1013}: in -> px -> rfg -> A

Ultimately, three parts are accepted. Adding up the x, m, a, and s rating for each of the accepted parts gives 7540 for the part with x=787, 4623 for the part with x=2036, and 6951 for the part with x=2127. Adding all of the ratings for all of the accepted parts gives the sum total of 19114.
Sort through all of the parts you've been given; what do you get if you add together all of the rating numbers for all of the parts that ultimately get accepted?

*/
fun main() {

    val inlineTestInput = """
px{a<2006:qkq,m>2090:A,rfg}
pv{a>1716:R,A}
lnx{m>1548:A,A}
rfg{s<537:gd,x>2440:R,A}
qs{s>3448:A,lnx}
qkq{x<1416:A,crn}
crn{x>2662:A,R}
in{s<1351:px,qqz}
qqz{s>2770:qs,m<1801:hdj,R}
gd{a>3333:R,R}
hdj{m>838:A,pv}

{x=787,m=2655,a=1222,s=2876}
{x=1679,m=44,a=2067,s=496}
{x=2036,m=264,a=79,s=2244}
{x=2461,m=1339,a=466,s=291}
{x=2127,m=1623,a=2188,s=1013}
"""

    data class RuleCondition(val attr: Int = 0, val isSmaller: Boolean = false, val value: Int = 0, val follow: String)

    fun createMap(input: List<String>): Pair<Int, Map<String, List<RuleCondition>>> {
        val ruleMap = HashMap<String, List<RuleCondition>>()
        var p = 0
        do {
            "([a-z]+)\\{(.*)}".toRegex().matchEntire(input[p])?.destructured
                ?.let { (id, rules) ->
                    val rulesList = rules.split(",").map {
                        if (it.contains("<")) {
                            val (attr, rest) = it.split("<")
                            val (value, branchId) = rest.split(":")
                            RuleCondition("xmas".indexOf(attr), true, value.toInt(), branchId)
                        } else if (it.contains(">")) {
                            val (attr, rest) = it.split(">")
                            val (value, branchId) = rest.split(":")
                            RuleCondition("xmas".indexOf(attr), false, value.toInt(), branchId)
                        } else {
                            RuleCondition(follow = it)
                        }
                    }
                    ruleMap[id] = rulesList
                }
            if (input[p].isBlank()) break
            p++
        } while (true)
        return p to ruleMap
    }

    fun part1(input: List<String>): Int {
        val (p, ruleMap) = createMap(input)
        var sum = 0
        for (pp in p + 1 until input.size) {
            val attrs =
                input[pp].split("=", ",", "x", "m", "a", "s", "{", "}").filter { it.isNotBlank() }.map { it.toInt() }
                    .toIntArray()
            var ruleId = "in"
            while (ruleId != "A" && ruleId != "R") {
                val conditions = ruleMap[ruleId]!!
                for (cond in conditions) {
                    if (cond.isSmaller) {
                        if (attrs[cond.attr] < cond.value) {
                            ruleId = cond.follow
                            break
                        }
                    } else {
                        if (attrs[cond.attr] > cond.value) {
                            ruleId = cond.follow
                            break
                        }
                    }
                }
            }
            if (ruleId == "A") {
                sum += attrs.sum()
            }
        }
        return sum
    }

    fun recurse(ruleMap: Map<String, List<RuleCondition>>, ranges: List<IntRange>, ruleId: String): Long {
        if (ruleId == "A") {
            return ranges.map { (it.last - it.first + 1).coerceAtLeast(0).toLong() }.reduce(Long::times)
        }
        if (ruleId == "R") return 0
        var result = 0L
        val conditions = ruleMap[ruleId]!!
        var currRanges = ranges
        for (cond in conditions) {
            if (cond.value != 0) {
                if (cond.isSmaller) {
                    val subRange = currRanges.mapIndexed { index, intRange ->
                        if (index != cond.attr) intRange else
                            IntRange(intRange.first, intRange.last.coerceAtMost(cond.value - 1))
                    }
                    result += recurse(ruleMap, subRange, cond.follow)
                    currRanges = currRanges.mapIndexed { index, intRange ->
                        if (index != cond.attr) intRange else
                            IntRange(intRange.first.coerceAtLeast(cond.value), intRange.last)
                    }
                } else {
                    val subRange = currRanges.mapIndexed { index, intRange ->
                        if (index != cond.attr) intRange else
                            IntRange(intRange.first.coerceAtLeast(cond.value + 1), intRange.last)
                    }
                    result += recurse(ruleMap, subRange, cond.follow)
                    currRanges = currRanges.mapIndexed { index, intRange ->
                        if (index != cond.attr) intRange else
                            IntRange(intRange.first, intRange.last.coerceAtMost(cond.value))
                    }
                }
            } else {
                result += recurse(ruleMap, currRanges, cond.follow)
                break
            }
        }
        return result
    }

    fun part2(input: List<String>): Long {
        val (_, ruleMap) = createMap(input)
        val ranges = listOf(IntRange(1, 4000), IntRange(1, 4000), IntRange(1, 4000), IntRange(1, 4000))
        return recurse(ruleMap, ranges, "in")
    }

    // test if implementation meets criteria from the description, like:
    val testInput = inlineTestInput.trim().reader().readLines()
    //val testInput = readInput("aoc2023/Day19_test")
    val testInputPart1Result = part1(testInput)
    println("Part 1 Test: $testInputPart1Result")
    val testInputPart2Result = part2(testInput)
    println("Part 2 Test: $testInputPart2Result")
    check(testInputPart1Result == 19114)
    check(testInputPart2Result == 167409079868000L)

    val input = readInput("aoc2023/Day19")
    part1(input).println()
    part2(input).println()
}
